require 'rubygems'
require 'bundler'
require 'net/http'
require 'json'
require 'logger'

Bundler.require

module WEBrick
  class HTTPResponse
    def create_error_page
      @body = { :error => HTMLUtils::escape(@reason_phrase) }.to_json
    end
  end
end

class Logger
  def internal(error)
    error '[!] INTERNAL SERVER ERROR - backtrace:'
    error error.inspect
    error error.backtrace
    error '[!] INTERNAL SERVER ERROR - end of backtrace.'
  end
end

module Git
  class Lib
    def remote_add(name, url, opts = {})
      arr_opts = ['add']
      arr_opts << '-f' if opts[:with_fetch] || opts[:fetch]
      arr_opts << '-t' << opts[:track] if opts[:track]
      arr_opts << '--mirror=fetch'if opts[:mirror]
      arr_opts << '--'
      arr_opts << name
      arr_opts << url

      command('remote', arr_opts)
    end
  end
end

class OctoLab < Sinatra::Base

  GITHUB_SECRET = ENV['GITHUB_SECRET'].to_s
  GITHUB_TOKEN  = ENV['GITHUB_TOKEN'].to_s
  GITLAB_TOKEN  = ENV['GITLAB_TOKEN'].to_s
  GITLAB_CI_URL = ENV['GITLAB_CI_URL'].to_s
  GIT_DATA_PATH = ENV['GIT_DATA_PATH'].to_s
  PROCESS_UID   = ENV['PROCESS_UID'].to_s
  PROCESS_GID   = ENV['PROCESS_GID'].to_s
  LOG_DIR_PATH  = ENV['LOG_DIR_PATH'].to_s
  LOG_WEB_AGE   = ENV['LOG_WEB_AGE'].to_s
  LOG_WEB_SIZE  = ENV['LOG_WEB_SIZE'].to_s
  LOG_GIT_AGE   = ENV['LOG_GIT_AGE'].to_s
  LOG_GIT_SIZE  = ENV['LOG_GIT_SIZE'].to_s
  ROOT_DIR      = File.realpath(__FILE__.split('/')[0..-3].join('/'))

  class HTTPError < StandardError
  end

  class TriggerError < StandardError
  end
  
  configure do
    set :root, __FILE__.split('/')[0..-3].join('/')
    set :server, 'thin'
    set :run, true
    set :haml, { :ugly => true }
    set :clean_trace, true
    enable :logging
    
    
    if settings.production?
      $log_app = Logger.new(LOG_DIR_PATH + '/app.log', LOG_WEB_AGE, LOG_WEB_SIZE)
      $log_git = Logger.new(LOG_DIR_PATH + '/git.log', LOG_GIT_AGE, LOG_GIT_SIZE)
      $log_app.level = Logger::DEBUG
      $log_git.level = Logger::DEBUG
    else
      $log_app = Logger.new(STDOUT)
      $log_git = Logger.new(STDOUT)
      $log_app.level = Logger::DEBUG
      $log_git.level = Logger::DEBUG
    end
  end

  before do
    @client ||= Octokit::Client.new(:access_token => GITHUB_TOKEN)
  end

  post '/events' do
    begin
      verify_signature(request.body.read)
      request.body.rewind
      payload = JSON.parse(request.body.read)
      $log_app.debug JSON.pretty_generate(payload) if settings.development?
      $log_app.info '[*] Received event: ' + request.env['HTTP_X_GITHUB_EVENT'] if request.env['HTTP_X_GITHUB_EVENT']
      
      case request.env['HTTP_X_GITHUB_EVENT']
      when 'pull_request'
        git_fetch_github(payload['pull_request']['base']['repo']['full_name'])
        if payload['action'] =~ /(re)*opened/
          ENV['OCTOLAB_COMMITS'] = (ENV['OCTOLAB_COMMITS'].to_s.split(',') + [payload['pull_request']['head']['sha']]).join(',')
          process_pull_request(payload['pull_request'])
        end
      when 'push'
        git_fetch_github(payload['repository']['full_name'])
        process_push(payload)
      end
    rescue
      $log_app.internal($!)
      halt 500, { :error => 'Processing failed' }.to_json
    end
  end

  post '/builds' do
    begin
      payload = JSON.parse(request.body.read)
      $log_app.debug JSON.pretty_generate(payload) if settings.development?

      case request.env['HTTP_X_GITLAB_EVENT']
      when 'Build Hook'
        local_name    = payload['project_name'].gsub(/\s\/\s/, '/')
        commit        = payload['sha']
        git_dir       = GIT_DATA_PATH + '/' + local_name + '.git'
        git           = Git.open(git_dir, {:log => $log_git, :repository => git_dir, :working_directory => nil })
        remote_name   = git.remotes.select { |i| i.name == 'github' }.first.url.split(':').last.gsub(/\.git$/, '')
        build_id      = payload['build_id'].to_s
        commit_status = payload['commit']['status']
        build_status  = payload['build_status']
        build_status  = 'failure' if payload['build_status'] == 'canceled' or payload['build_status'] == 'failed'
        description   = 'The build is in progress...'
        description   = 'The build succeeded!'  if build_status == 'success'
        description   = 'The build has failed!' if build_status == 'failure'
        options       = {
                          :state        => build_status,
                          :target_url   => GITLAB_CI_URL + '/' + remote_name + '/builds/' + build_id,
                          :description  => description,
                          :context      => 'continuous-integration/gitlab'
                        }

        $log_app.info '[*] Build status changed: ' + remote_name + ': ' + commit + ': ' + build_status

        if ENV['OCTOLAB_COMMITS'].to_s.split(',').include?(commit)
          $log_app.info '[*] Sending status update: ' + remote_name + ': ' + commit + ': ' + build_status
          if build_status =~ /pending/
            @client.create_status(remote_name, commit, build_status, options)
          elsif build_status =~ /failure/ or (build_status =~ /success/ and commit_status =~ /success/)
            @client.create_status(remote_name, commit, build_status, options)
            ENV['OCTOLAB_COMMITS'] = ENV['OCTOLAB_COMMITS'].to_s.split(',').delete_if { |i| i == commit }.join(',')
          end
        end
        
        return
      end
    rescue
      $log_app.internal($!)
      halt 500, { :error => 'Processing failed' }.to_json
    end
  end

  helpers do
    def logger
      request.logger
    end
    
    def verify_signature(body)
      signature = 'sha1=' + OpenSSL::HMAC.hexdigest(OpenSSL::Digest.new('sha1'), GITHUB_SECRET, body)
      return halt 500, { :error => "Signatures didn't match!" }.to_json unless Rack::Utils.secure_compare(signature, request.env['HTTP_X_HUB_SIGNATURE'])
    end
    
    def git_fetch_github(full_name)
      $log_app.info '[*] Fetching: ' + full_name

      Git.configure do |config|
        config.git_ssh = ROOT_DIR + '/helpers/ssh'
      end

      git_dir = GIT_DATA_PATH + '/' + full_name + '.git'
      git = Git.open(git_dir, { :log => $log_git, :repository => git_dir, :working_directory => nil })
      
      if git.remotes.length == 0
        git.add_remote('github', 'git@github.com:' + full_name + '.git', :mirror => true)
      end
      
      git.remote('github').fetch
    end
    
    def get_project(name)
      begin
        url           = GITLAB_CI_URL + '/api/v3/projects'
        uri           = URI(url)
        http          = Net::HTTP.new(uri.host, uri.port)
        http.use_ssl  = true
        headers       = { 'PRIVATE-TOKEN' => GITLAB_TOKEN }
        path          = uri.path.empty? ? '/' : uri.path
        code          = http.head(path, headers).code.to_i

        if (code >= 200 && code < 300)
          response = http.get(uri.path, headers)
          JSON.parse(response.body).select { |i| i['path_with_namespace'] == name }.first
        else
          raise HTTPError, 'Error code: ' + code.to_s + ' URL: ' + url
        end
      rescue
        raise $!
      end
    end
    
    def get_trigger_token(id)
      begin
        uri           = URI(GITLAB_CI_URL + '/api/v3/projects/' + id.to_s + '/triggers')
        http          = Net::HTTP.new(uri.host, uri.port)
        http.use_ssl  = true
        headers       = { 'PRIVATE-TOKEN' => GITLAB_TOKEN }
        path          = uri.path.empty? ? '/' : uri.path
        code          = http.head(path, headers).code.to_i

        if (code >= 200 && code < 300)
          response = http.get(uri.path, headers)
          res = JSON.parse(response.body).select { |i| i['deleted_at'] == nil }.first
          if res == nil
            raise HTTPError, 'Unable to obtain build trigger token for project id: ' + id.to_s
          end
          res
        else
          raise HTTPError, 'Error code: ' + code.to_s + ' URL: ' + url
        end
      rescue
        raise $!
      end
    end
    
    def trigger_build(remote_name, ref)
      begin
        project       = get_project(remote_name)
        token         = get_trigger_token(project['id'])['token']
        url           = GITLAB_CI_URL + '/api/v3/projects/' + project['id'].to_s + '/trigger/builds'
        uri           = URI(url)
        response      = Net::HTTP.post_form(uri, 'token' => token, 'ref' => ref)
        
        if (response.code.to_i >= 200 && response.code.to_i < 300)
          JSON.parse(response.body)
        elsif response.code.to_i == 400
          raise TriggerError, JSON.parse(response.body)['message']
        else
          raise HTTPError, 'Error code: ' + response.code.to_s + ' URL: ' + url
        end
      rescue
        raise $!
      end
    end
    
    def process_pull_request(pull_request)
      Thread.new do
        begin
          remote_name   = pull_request['base']['repo']['full_name']
          commit        = pull_request['head']['sha']
          ref           = pull_request['head']['ref']
          build_status  = 'pending'
          description   = 'The build is in progress...'
          options       = {
                          :state        => build_status,
                          :description  => description,
                          :context      => 'continuous-integration/gitlab'
                        }

          $log_app.info '[*] Processing pull request: ' + remote_name + ': ' + commit
          @client.create_status(remote_name, commit, build_status, options)
          trigger_build(remote_name, ref)
        rescue Octokit::Unauthorized
          $log_app.error '[!] ' + $!.message
        rescue HTTPError
          $log_app.error $!
        rescue TriggerError
          $log_app.error $!
        rescue
          $log_app.error '[!] Unable to process request for: ' + remote_name + ' ref: ' + ref
          $log_app.internal($!)
        ensure
          build_status  = 'failure'
          description   = 'The build has failed!'
          options       = {
                          :state        => build_status,
                          :description  => description,
                          :context      => 'continuous-integration/gitlab'
                        }

          @client.create_status(remote_name, commit, build_status, options)
        end
      end
    end
    
    def process_push(payload)
      Thread.new do
        begin
          remote_name   = payload['repository']['full_name']
          ref           = payload['ref'].split("/").last
          
          trigger_build(remote_name, ref)
        rescue HTTPError
          $log_app.error $!
        rescue TriggerError
          $log_app.error $!
        rescue
          $log_app.error '[!] Unable to process request for: ' + remote_name + ' ref: ' + ref
          $log_app.internal($!)
        end
      end
    end   
  end
end
